xquery version "3.1";

module namespace studienausgabe="https://mollenhauer-edition.de/ns/conversions/studienausgabe";

(:~
 : Convert Module Studienausgae
 : *******************************************************************************************
 : This module contains functions to convert KMG-TEI to a format fitting for the Studienausgabe.
 :
 : @version 1.4 (2024-05-27)
 : @status stable
 : @author Max-Ferdinand Zeterberg
 :)

import module namespace bibref="https://mollenhauer-edition.de/ns/conversions/bibref" at "bibref.xqm";
import module namespace citations="https://mollenhauer-edition.de/ns/conversions/citations" at "citations.xqm";
import module namespace selection = "https://mollenhauer-edition.de/ns/conversions/selection" at "selection.xqm";
import module namespace simplify="https://mollenhauer-edition.de/ns/conversions/simplify" at "simplify.xqm";

declare namespace tei="http://www.tei-c.org/ns/1.0";

declare namespace functx = "http://www.functx.com";

declare function functx:is-node-in-sequence-deep-equal
  ( $node as node()? ,
    $seq as node()* )  as xs:boolean {

   some $nodeInSeq in $seq satisfies deep-equal($nodeInSeq,$node)
 } ;

declare function functx:distinct-deep
  ( $nodes as node()* )  as node()* {

    for $seq in (1 to count($nodes))
    return $nodes[$seq][not(functx:is-node-in-sequence-deep-equal(
                          .,$nodes[position() < $seq]))]
 } ;

declare function functx:is-value-in-sequence
  ( $value as xs:anyAtomicType? ,
    $seq as xs:anyAtomicType* )  as xs:boolean {

   $value = $seq
 } ;

declare function functx:sort
  ( $seq as item()* )  as item()* {

   for $item in $seq
   order by $item
   return $item
 } ;

(:~
 : This function produces the full bibilographical reference for one title as a tei:bibl. If the data cannot be retrieved, the error-element will be forwarded.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $bibliographic-ID KMG-Number
 : @return bibliographical reference as string
 :)

declare function studienausgabe:fullreference-kmg($bibliographic-ID as xs:string, $type as xs:string?) as node()
{
    let $kmg-ID := tokenize($bibliographic-ID, ':')[2]
    let $listBibl := bibref:make-listbibl($kmg-ID)
    let $biblStruct := $listBibl/*[1]
    return
        if ($biblStruct[self::error]) then
            element error {
                attribute n {5},
                attribute bibliographical-ID {$bibliographic-ID},
                $biblStruct/text()
            }
        else
            bibref:fullreference-TEI($biblStruct, $type)
};

(:~
 : This function produces the full bibilographical reference for one title as a tei:bibl. If the data cannot be retrieved from Zotero, the error-element will be forwarded.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $zotero-Id ID of the Zotero-Item
 : @param $type Typing for Bibls
 : @return bibliographical reference as string
 :)

declare function studienausgabe:fullreference-zotero($bibliographic-ID as xs:string, $type as xs:string?) as node()
{
    let $zotero-ID := tokenize($bibliographic-ID, ':')[2]
    let $zotero-data := selection:get-zotero-data($zotero-ID)
    return
        if ($zotero-data[self::error]) then
            element error {
                attribute n {$zotero-data/@n},
                attribute bibliographical-ID {$bibliographic-ID},
                $zotero-data/text()
            }
        else
            let $references := $zotero-data//fn:string[@key="bib"]
            let $xml := fn:parse-xml($references)
            let $referencesXML := $xml//div[@class = 'csl-entry']/node()
            return
                element {QName('http://www.tei-c.org/ns/1.0', 'bibl')} {
                    if ($type) then
                        attribute type {$type}
                    else
                        (),
                    for $reference in $referencesXML
                    return
                        typeswitch ($reference)
                            case element(i)
                                return
                                    element {QName('http://www.tei-c.org/ns/1.0', 'hi')} {
                                        attribute rend {'italics'},
                                        $reference/node()
                                    }
                            default
                                return
                                    ($reference)
                }
};

(:~
 : This function retrieves the data bibliographical data for citations.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $bibls tei:bibl
 : @param $type Typing for Bibls
 : @return bibl or error elements
 :)

declare function studienausgabe:all-references($bibliographicIDs as xs:string*, $type as xs:string) as element()*
{

    for $bibliographicID in $bibliographicIDs
    return
        if (starts-with($bibliographicID, 'zotero:')) then
            studienausgabe:fullreference-zotero($bibliographicID, $type)
        else if (starts-with($bibliographicID, 'kmglit:')) then
            studienausgabe:fullreference-kmg($bibliographicID, $type)
        else
            element error {
                attribute n {4},
                attribute bibliographical-ID {$bibliographicID},
                'Bibliographical-ID ' || $bibliographicID || ' does not exist'
            }
};

(:~
 : This function retrieves the bibliographical ID from a bibl-element.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $bibls tei:bibl
 : @return bibliographical ID
 :)

declare function studienausgabe:get-all-IDs($bibls as node()*)
{
    for $bibl in $bibls
    let $biblID := citations:get-bibliographic-id($bibl)
    return
        $biblID
};

(:~
 : This function retrieves the bibliographical data for all citations
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $doc A TEI-Document
 : @param $fassung The Fassung that should be used
 : @return bibliographical data as JSON
 :)
declare function studienausgabe:all-citations($doc as node()*)
{
    let $commentaryBibls := $doc//tei:text[@type='edition']//tei:bibl[@type='commentary']
    let $commentaryBiblIDs := studienausgabe:get-all-IDs($commentaryBibls)
    let $distinctCommentaryBiblIDs := distinct-values($commentaryBiblIDs)
    let $commentaryreferences := studienausgabe:all-references($distinctCommentaryBiblIDs, "stellenkommentar")
    let $werkkommentarBibls := $doc//tei:text[@type='commentary']//tei:bibl
    let $werkkommentarBiblIDs := studienausgabe:get-all-IDs($werkkommentarBibls)
    let $distinctWerkkommentarBiblIDs := distinct-values($werkkommentarBiblIDs)
    let $werkkommentarreferences := studienausgabe:all-references($distinctWerkkommentarBiblIDs, "werkkommentar")
    let $normalBibls := $doc//tei:text[@type='edition']//tei:bibl[not(@type)]
    let $KMGBibls := $doc//tei:text[@type='edition']//tei:bibl[@type='KMG']
    let $normalBiblIDs := studienausgabe:get-all-IDs($normalBibls)
    let $distinctNormalBiblIDs := distinct-values($normalBiblIDs)
    let $normalReferences := studienausgabe:all-references($distinctNormalBiblIDs, "KM")
    let $KMGBiblIDs := (
        for $KMGBiblID in studienausgabe:get-all-IDs($KMGBibls)
        return
            if (functx:is-value-in-sequence($KMGBiblID, $distinctNormalBiblIDs)) then
                ()
            else
                $KMGBiblID
        )
    let $distinctKMGBiblIDs := distinct-values($KMGBiblIDs)
    let $KMGReferences := studienausgabe:all-references($distinctKMGBiblIDs, "KMG")
    return
        (
            element {QName('http://www.tei-c.org/ns/1.0', 'listBibl')} {
                attribute type {'KM'},
                functx:sort(($normalReferences, $KMGReferences))
            },
            element {QName('http://www.tei-c.org/ns/1.0', 'listBibl')} {
                attribute type {'Stellenkommentare'},
                functx:sort($commentaryreferences)
            },
            element {QName('http://www.tei-c.org/ns/1.0', 'listBibl')} {
                attribute type {'Werkkommentar'},
                functx:sort($werkkommentarreferences)
            }
        )
};

(:~
 : Default return for studienausgabe:new-div-with-back.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $doc A node
 : @param $listBibl Bibliographical data as a sequence of tei:listBibl
 : @return $doc A node
 :)
declare function studienausgabe:default-new-div-with-back($node as node()*, $listBibl as node()*)
{
    try {
        element {QName($node/namespace-uri(), $node/name())} {
            $node/@*,
            studienausgabe:new-div-with-back($node/node(), $listBibl)
        }
    } catch * {
        studienausgabe:new-div-with-back($node/node(), $listBibl)
    }
};

(:~
 : This function inserts new tei:div in a document where tei:text[@type='edition']/tei:back exists.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $doc A TEI-Document
 : @param $listBibl Bibliographical data as a sequence of tei:listBibl
 : @return updated TEI-Document
 :)
declare function studienausgabe:new-div-with-back($doc as node()*, $listBibl as node()*) as node()*
{
    for $node in $doc
        return
            typeswitch ($node)
                case text()
                    return
                        ($node)
                case processing-instruction()
                    return
                        ($node)
                case element(tei:back)
                    return
                        if ($node/ancestor::tei:text[@type='edition']) then
                            element {QName('http://www.tei-c.org/ns/1.0', 'back')} {
                                studienausgabe:new-div-with-back($node/node(), $listBibl),
                                if ($node/ancestor::tei:text//tei:note[@type='commentary']) then
                                    element {QName('http://www.tei-c.org/ns/1.0', 'div')}{
                                        attribute type {'endnoten'}
                                    }
                                else (),
                                element {QName('http://www.tei-c.org/ns/1.0', 'div')} {
                                    attribute type {'LiteraturSTA'},
                                    $listBibl[@type='Stellenkommentare'],
                                    $listBibl[@type='KM']
                                }
                            }
                        else
                            element {QName('http://www.tei-c.org/ns/1.0', 'back')} {
                                studienausgabe:new-div-with-back($node/node(), $listBibl),
                                element {QName('http://www.tei-c.org/ns/1.0', 'div')} {
                                    attribute type {'LiteraturSTAKommentar'},
                                    $listBibl[@type='Werkkommentar']
                                }
                            }
                default
                    return
                        studienausgabe:default-new-div-with-back($node, $listBibl)
};

(:~
 : Default return for studienausgabe:new-div-without-back.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $doc A node
 : @param $listBibl Bibliographical data as a sequence of tei:listBibl
 : @return $doc A node
 :)
declare function studienausgabe:default-new-div-without-back($node as node()*, $listBibl as node()*)
{
    try {
        element {QName($node/namespace-uri(), $node/name())} {
            $node/@*,
            studienausgabe:new-div-without-back($node/node(), $listBibl)
        }
    } catch * {
        studienausgabe:new-div-without-back($node/node(), $listBibl)
    }
};

(:~
 : This function inserts new tei:div in a document where tei:text[@type='edition']/tei:back does not exists.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $doc A TEI-Document
 : @param $listBibl Bibliographical data as a sequence of tei:listBibl
 : @return updated TEI-Document
 :)
declare function studienausgabe:new-div-without-back($doc as node()*, $listBibl as node()*) as node()*
{
    for $node in $doc
        return
            typeswitch ($node)
                case text()
                    return
                        ($node)
                case processing-instruction()
                    return
                        ($node)
                case element(tei:text)
                    return
                        if ($node[@type='edition']) then
                            element {QName('http://www.tei-c.org/ns/1.0', 'text')} {
                                attribute type {$node/@type},
                                attribute xml:id {$node/@xml:id},
                                attribute xml:lang {$node/@xml:lang},
                                studienausgabe:new-div-without-back($node/node(), $listBibl),
                                element {QName('http://www.tei-c.org/ns/1.0', 'back')} {
                                    if ($node//tei:note[@type='commentary']) then
                                        element {QName('http://www.tei-c.org/ns/1.0', 'div')}{
                                            attribute type {'endnoten'}
                                        }
                                    else (),
                                    element {QName('http://www.tei-c.org/ns/1.0', 'div')} {
                                        attribute type {'LiteraturSTA'},
                                        $listBibl[@type='Stellenkommentare'],
                                        $listBibl[@type='KM']
                                    }
                                }
                            }
                        else
                            studienausgabe:default-new-div-without-back($node, $listBibl)
                case element(tei:back)
                    return
                        element {QName('http://www.tei-c.org/ns/1.0', 'back')} {
                            studienausgabe:new-div-without-back($node/node(), $listBibl),
                            element {QName('http://www.tei-c.org/ns/1.0', 'div')} {
                                attribute type {'LiteraturSTAKommentar'},
                                $listBibl[@type='Werkkommentar']
                            }
                        }
                default
                    return
                        studienausgabe:default-new-div-without-back($node, $listBibl)
};

(:~
 : This function updates a TEI-document with new tei:listBibl.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $doc A TEI-Document
 : @param $listBibl Bibliographical data as a sequence of tei:listBibl
 : @return updated TEI-Document
 :)
declare function studienausgabe:new-doc($doc as node()*, $listBibl as node()*) as node()*
{
    if ($doc//tei:text[@type='edition']/tei:back) then
        studienausgabe:new-div-with-back($doc, $listBibl)
    else
        studienausgabe:new-div-without-back($doc, $listBibl)
};
